import asyncio
import logging
from functools import wraps
from collections.abc import Set
from threading import Lock
from typing import Callable, Iterable


__all__ = 'create_task', 'iscoroutinepartial', 'shield', 'CallbackCollection'

log = logging.getLogger(__name__)


def iscoroutinepartial(fn):
    """
    Function returns True if function is a partial instance of coroutine.
    See additional information here_.

    :param fn: Function
    :return: bool

    .. _here: https://goo.gl/C0S4sQ

    """

    while True:
        parent = fn

        fn = getattr(parent, 'func', None)

        if fn is None:
            break

    return asyncio.iscoroutinefunction(parent)


def create_task(func, *args, loop=None, **kwargs):
    loop = loop or asyncio.get_event_loop()

    if iscoroutinepartial(func):
        return loop.create_task(func(*args, **kwargs))

    def run(future):
        if future.done():
            return

        try:
            future.set_result(func(*args, **kwargs))
        except Exception as e:
            future.set_exception(e)

        return future

    future = loop.create_future()
    loop.call_soon(run, future)
    return future


def shield(func):
    """
    Simple and useful decorator for wrap the coroutine to `asyncio.shield`.
    """

    async def awaiter(future):
        return await future

    @wraps(func)
    def wrap(*args, **kwargs):
        return wraps(func)(awaiter)(asyncio.shield(func(*args, **kwargs)))

    return wrap


class CallbackCollection(Set):
    __slots__ = '__callbacks', '__lock'

    def __init__(self):
        self.__callbacks = set()
        self.__lock = Lock()

    def add(self, callback: Callable):
        if self.is_frozen:
            raise RuntimeError('Collection frozen')
        if not callable(callback):
            raise ValueError("Callback is not callable")

        with self.__lock:
            self.__callbacks.add(callback)

    def remove(self, callback: Callable):
        if self.is_frozen:
            raise RuntimeError('Collection frozen')

        with self.__lock:
            self.__callbacks.remove(callback)

    def clear(self):
        if self.is_frozen:
            raise RuntimeError('Collection frozen')

        with self.__lock:
            self.__callbacks.clear()

    @property
    def is_frozen(self) -> bool:
        return isinstance(self.__callbacks, frozenset)

    def freeze(self):
        if self.is_frozen:
            raise RuntimeError("Collection already frozen")

        with self.__lock:
            self.__callbacks = frozenset(self.__callbacks)

    def unfreeze(self):
        if not self.is_frozen:
            raise RuntimeError("Collection is not frozen")

        with self.__lock:
            self.__callbacks = set(self.__callbacks)

    def __contains__(self, x: object) -> bool:
        return x in self.__callbacks

    def __len__(self) -> int:
        return len(self.__callbacks)

    def __iter__(self) -> Iterable[Callable]:
        return iter(self.__callbacks)

    def __bool__(self):
        return bool(self.__callbacks)

    def __copy__(self):
        instance = self.__class__()

        with self.__lock:
            for cb in self.__callbacks:
                instance.add(cb)

        if self.is_frozen:
            instance.freeze()

        return instance

    def __call__(self, *args, **kwargs):
        with self.__lock:
            for cb in self.__callbacks:
                try:
                    cb(*args, **kwargs)
                except Exception:
                    log.exception('Callback error')



from typing import Union

import aiormq

from logging import getLogger
from warnings import warn

from .exchange import Exchange, ExchangeType
from .queue import Queue
from .types import TimeoutType
from .channel import Channel
from .robust_queue import RobustQueue
from .robust_exchange import RobustExchange


log = getLogger(__name__)


class RobustChannel(Channel):
    """ Channel abstraction """

    QUEUE_CLASS = RobustQueue
    EXCHANGE_CLASS = RobustExchange

    def __init__(self, connection, channel_number: int = None,
                 publisher_confirms: bool = True, on_return_raises=False):
        """

        :param connection: :class:`aio_pika.adapter.AsyncioConnection` instance
        :param loop:
            Event loop (:func:`asyncio.get_event_loop()` when :class:`None`)
        :param future_store: :class:`aio_pika.common.FutureStore` instance
        :param publisher_confirms:
            False if you don't need delivery confirmations
            (in pursuit of performance)
        """

        super().__init__(
            connection=connection,
            channel_number=channel_number,
            publisher_confirms=publisher_confirms,
            on_return_raises=on_return_raises,
        )

        self._exchanges = dict()
        self._queues = dict()
        self._prefetch_count = 0
        self._prefetch_size = 0
        self._global_ = False

    async def reopen(self):
        await super().reopen()
        await self.restore()

    async def restore(self):
        await self.set_qos(
            prefetch_count=self._prefetch_count,
            prefetch_size=self._prefetch_size,
            global_=self._global_
        )

        for exchange in self._exchanges.values():
            await exchange.restore(self)

        for queue in self._queues.values():
            await queue.restore(self)

    def _on_channel_close(self, exc: Exception):
        if exc:
            log.exception(
                "Robust channel %r has been closed.", self, exc_info=exc
            )
        else:
            log.debug("Robust channel %r has been closed.", self)

    async def initialize(self, timeout: TimeoutType = None) -> None:
        await super().initialize(timeout)
        self.add_close_callback(self._on_channel_close)

    async def set_qos(self, prefetch_count: int = 0, prefetch_size: int = 0,
                      global_: bool = False, timeout: TimeoutType = None,
                      all_channels: bool = None):
        if all_channels is not None:
            warn('Use "global_" instead of "all_channels"', DeprecationWarning)
            global_ = all_channels

        self._prefetch_count = prefetch_count
        self._prefetch_size = prefetch_size
        self._global_ = global_

        return await super().set_qos(
            prefetch_count=prefetch_count,
            prefetch_size=prefetch_size,
            global_=global_,
            timeout=timeout,
        )

    async def declare_exchange(
        self, name: str, type: Union[ExchangeType, str] = ExchangeType.DIRECT,
        durable: bool = None, auto_delete: bool = False,
        internal: bool = False, passive: bool = False, arguments: dict = None,
        timeout: TimeoutType = None, robust: bool = True
    ) -> Exchange:
        exchange = await super().declare_exchange(
            name=name, type=type, durable=durable, auto_delete=auto_delete,
            internal=internal, passive=passive, arguments=arguments,
            timeout=timeout,
        )

        if not internal and robust:
            self._exchanges[name] = exchange

        return exchange

    async def exchange_delete(self, exchange_name: str,
                              timeout: TimeoutType = None, if_unused=False,
                              nowait=False) -> aiormq.spec.Exchange.DeleteOk:

        result = await super().exchange_delete(
            exchange_name=exchange_name, timeout=timeout,
            if_unused=if_unused, nowait=nowait
        )

        self._exchanges.pop(exchange_name, None)

        return result

    async def declare_queue(self, name: str = None, *, durable: bool = None,
                            exclusive: bool = False, passive: bool = False,
                            auto_delete: bool = False, arguments: dict = None,
                            timeout: TimeoutType = None,
                            robust: bool = True) -> Queue:

        queue = await super().declare_queue(
            name=name, durable=durable, exclusive=exclusive,
            passive=passive, auto_delete=auto_delete,
            arguments=arguments, timeout=timeout,
        )

        if robust:
            self._queues[name] = queue

        return queue

    async def queue_delete(self, queue_name: str, timeout: TimeoutType = None,
                           if_unused: bool = False, if_empty: bool = False,
                           nowait: bool = False):
        result = await super().queue_delete(
            queue_name=queue_name, timeout=timeout,
            if_unused=if_unused, if_empty=if_empty, nowait=nowait
        )

        self._queues.pop(queue_name, None)
        return result


__all__ = ('RobustChannel',)



from . import patterns
from . import pool
from .connection import Connection, connect
from .channel import Channel
from .exchange import Exchange, ExchangeType
from .message import Message, IncomingMessage, DeliveryMode
from .queue import Queue
from .robust_connection import connect_robust, RobustConnection
from .robust_channel import RobustChannel
from .robust_exchange import RobustExchange
from .robust_queue import RobustQueue
from .exceptions import AMQPException, MessageProcessError
from .version import (
    __author__,
    __version__,
    author_info,
    package_info,
    package_license,
    version_info,
)


__all__ = (
    '__author__',
    '__version__',
    'author_info',
    'connect',
    'connect_robust',
    'package_info',
    'package_license',
    'patterns',
    'pool',
    'version_info',
    'AMQPException',
    'Channel',
    'Connection',
    'DeliveryMode',
    'Exchange',
    'ExchangeType',
    'IncomingMessage',
    'Message',
    'MessageProcessError',
    'Queue',
    'RobustChannel',
    'RobustConnection',
    'RobustExchange',
    'RobustQueue',
)
